"""Logging configuration, customization, and helpers.

"""

import logging


def prep_log(level, suppress_verbose_loggers=True):
    """Prepare logging

    Parameters
    ----------

    level :

    suppress_verbose_loggers : bool

        A lot of this code's dependencies have very verbose debug
        logging that is often not wanted when debugging more local
        issues.  This flag allows toggling suppression of those.

    Returns
    -------

    log :

    """
    log = logging.getLogger()
    log.setLevel(level)
    if log.hasHandlers():
        log.handlers.clear()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('[%(levelname)s] %(message)s (%(name)s)')
    handler.setFormatter(formatter)
    log.addHandler(handler)

    if suppress_verbose_loggers:
        dont_log_list = ['zeep.xsd.schema', 'zeep', 'zeep.transports',
                         'urllib3.connectionpool', 'zeep.wsdl.wsdl',
                         'zeep.xsd.visitor']
        for blisted in dont_log_list:
            logging.getLogger(blisted).disabled = True

    return log


def log_configuration(log, configs):
    """Logs some of the configuration parameters for current run

    """
    log.info('-----------------------------------')
    log.info("Detection threshold: " + str(configs['CONF_THRESHOLD']))
    log.info('Tracked classes: ' + str(configs['TRACKED_CLASS']))
    ipan, itilt, izoom = configs['INIT_POS']
    log.info(f'Initial position: {ipan} pan, {itilt} tilt, {izoom} zoom')

    if configs['RECORD']:
        log.info('Recording is turned ON')
        strg = configs['RECORD_FOLDER']
        log.info(f'Recordings will be stored in {strg}')
        strg = configs['TIMELAPSE_DELAY']
        log.info(f'{strg} seconds between timelapse frames.')
    else:
        log.info('Recording is turned OFF')

    if configs['RECORD_ONLY_DETECTIONS']:
        log.info('Record only detections: True')

    log.info('-----------------------------------')



"""Tools for creating a simple UI for apps using package

Mostly for use in included examples and related utilities.

"""

import cv2
import numpy as np

mouseX = 250
mouseY = 250
zoom_command = None


def orient_frame(frame, orientation):
    """Rotates the image frame based on orientation string

    Helper function to quickly re-orient the image frame from the
    camera based on the camera's orientation. 'up' is the default. The
    orientation is usually provided in the user config yaml file as
    the orientation is assumed to be fixed for any given run.

    """
    if orientation == 'left':
        frame = np.rot90(frame)
    elif orientation == 'down':
        frame = np.rot90(frame, 2)
    elif orientation == 'right':
        frame = np.rot90(frame, 3)

    # Returning a copy (as implemented below) solves a bug that keeps
    # one from drawing on the resultant frame but might also allay
    # other problems that spawn from the same source.  The underlying
    # bug might be in the opencv library.
    return frame.copy()


def mouse_callback(event, x, y, flags, param):
    """Callback function for mouse interface

    """

    global mouseX
    global mouseY
    global zoom_command

    zoom_command = None

    if event == cv2.EVENT_MOUSEMOVE:
        mouseX = x
        mouseY = y

    if event == cv2.EVENT_LBUTTONDOWN:
        zoom_command = 'i'
    elif event == cv2.EVENT_LBUTTONUP or event == cv2.EVENT_RBUTTONDOWN:
        zoom_command = 'o'


class UI_Handler():
    """Presents a simple UI for applications

    """

    def __init__(self, frame, window_name, scale_display=1.0):
        self.window_name = window_name

        width = frame.shape[1]
        height = frame.shape[0]
        display_width = int(width * scale_display)
        display_height = int(height * scale_display)

        cv2.namedWindow(self.window_name, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(self.window_name,
                         display_width,
                         display_height)

        cv2.imshow(self.window_name, frame)
        cv2.setMouseCallback(self.window_name, mouse_callback)

        self.zone = {'start': (int(.3*width), int(.3*height)),
                     'end': (int(.7*width), int(.7*height))}

    def update(self, frame, hud=True):

        if hud:
            cv2.rectangle(frame,
                          self.zone['start'],
                          self.zone['end'],
                          (255, 0, 0),
                          thickness=2)

        cv2.imshow(self.window_name, frame)
        key = cv2.waitKey(10)

        return key

    def read_mouse(self):

        if mouseX < self.zone['start'][0]:
            x_dir = -1
        elif mouseX > self.zone['end'][0]:
            x_dir = 1
        else:
            x_dir = 0

        if mouseY < self.zone['start'][1]:
            y_dir = 1
        elif mouseY > self.zone['end'][1]:
            y_dir = -1
        else:
            y_dir = 0

        return x_dir, y_dir, zoom_command

    def clean_up(self):
        cv2.destroyAllWindows()



"""Tools for control of pan-tilt-zoom functionality of PTZ IP camera

"""
import logging
import time

import numpy as np
from onvif import ONVIFCamera

from camml import draw

log = logging.getLogger(__name__)


def _check_zeroness(number):
    """Almost-zero check

    Checks if a number is very close to zero (within a window) and if
    it is makes it less close by placing it at the edge of that
    window.

    Parameters
    ----------
    number : float
        A number to be checked for closeness to zero

    Returns
    -------
    number : float
        The "fixed" number

    """
    e = .001

    if -e < number < e:
        return 0
    else:
        return number


class MotorController():
    """Base class for motor controllers

    """
    # pylint: disable='too-few-public-methods'

    def __init__(self,
                 pid_gains,
                 orientation,
                 example_frame):
        """Constructor for MotorController class

        """

        self.pid_gains = pid_gains
        self.orientation = orientation

        self.frame_width = example_frame.shape[1]
        self.frame_height = example_frame.shape[0]

        self.total_frame_pixels = self.frame_width * self.frame_height

    def _calc_errors(self,
                     target_lbox):
        """Calculate errors from box coordinates

        Given a labeled box that is the bounding box of a detected
        target, this function calculates an error for each of x and y
        (pan and tilt) based on the offset of the centroid of the
        bounding box from the center of the frame.

        """
        if target_lbox:
            x_c, y_c = draw.box_to_coords(target_lbox['box'],
                                          return_kind='center')
            ret = draw.box_to_coords(target_lbox['box'])
            self.box_x, self.box_y, self.box_width, self.box_height = ret
            x_err = self.frame_width/2 - x_c
            y_err = self.frame_height/2 - y_c

            # normalize errors
            x_err = x_err/self.frame_width
            y_err = y_err/self.frame_height
        else:
            x_err = 0.0
            y_err = 0.0

        return x_err, y_err

    def update(self, target_lbox, zoom_command):
        """Update the camera commands

        Generates commands for pan, tilt, and zoom given the current
        target_lbox (which is None when no detection) and the current
        zoom_command.

        Parameters
        ----------
        target_lbox :

        zoom_command : float
            Current zoom_command

        Returns
        -------
        x_velocity :

        y_velocity :

        zoom_command :

        """
        errors = self._calc_errors(target_lbox)
        x_err, y_err = errors

        if self.orientation == 'down':
            x_err = -x_err
            y_err = -y_err

        x_velocity = self._calc_command(x_err, self.pid_gains[0])
        y_velocity = self._calc_command(y_err, self.pid_gains[1])
        zoom_command = self._calc_zoom_command(target_lbox,
                                               x_err,
                                               y_err,
                                               zoom_command)

        log.debug('x_err: %.2f || y_err: %.2f', x_err, y_err)
        log.debug('x_vel: %.2f || y_vel: %.2f', x_velocity, y_velocity)
        log.debug('zoom_command: %.2f', zoom_command)

        return (x_velocity, y_velocity, zoom_command)

    def _ensure_command_in_bounds(self, command):
        """Util to force command is within necessary bounds

        """
        if command >= 1.0:
            command = 1.0
        if command <= -1.0:
            command = -1.0

        return command

    def _calc_command(self, err, k):
        """Implements actual controller math

        This basic implementation in the base class is a mere
        proportional controller.  If need PI, PID, or such, this
        method should be replaced in the child class.  It is not
        currently imagined that controllers outside the PID space will
        be pursued.

        """
        command = k * err
        command = self._ensure_command_in_bounds(command)
        return command

    def _calc_zoom_command(self, target_lbox, x_err, y_err, zoom_command):
        """Calculate the next zoom command

        Not implemented in the base class. But should be implemented
        to effect zoom control based on position/size of target_lbox.

        """
        raise NotImplementedError


class CalmMotorController(MotorController):
    """MotorController with calm movements/behaviors

    In certain instances, the desire is not for rapid and exacting
    tracking but for tracking that doesn't distract from the scene
    through quick movements and jitteriness.  This subclass is
    intended to be that sort of calm tracker.

    """

    def __init__(self,
                 pid_gains,
                 orientation,
                 example_frame):

        super().__init__(pid_gains,
                         orientation,
                         example_frame)

        self.zoom_stop_ratio = .6
        self.stop_range = .1

    def _calc_command(self, err, k):
        """Override controller command method

        The meat here is that the controller stops moving axis under
        control if error is inside of some range.  This range is
        currently hardcoded but maybe should be given as an argument
        to the constructor.

        """

        if np.abs(err) < self.stop_range:
            command = 0
        else:
            command = k * err

        command = self._ensure_command_in_bounds(command)

        return command

    def _calc_zoom_command(self, target_lbox, x_err, y_err, zoom_command):
        """Calculate the zoom command give pan/tilt errors

        This is where most of the behavior of the tracking is
        implemented.

        """

        if (x_err <= 0.5 and y_err <= 0.5
           and target_lbox):
            target_bb_pixels = self.box_width * self.box_height

            box_ratio = target_bb_pixels / self.total_frame_pixels
            log.debug('Ratio of box to whole %.2f', box_ratio)

            if box_ratio < .3:
                zoom_command = 1.0
            else:
                zoom_command = 0.0

            ratio = self.zoom_stop_ratio
            spans_much_width = self.box_width >= ratio * self.frame_width
            spans_much_height = self.box_height >= ratio * self.frame_height
            if spans_much_width or spans_much_height:
                zoom_command = 0.0

            margin = 20
            if ((self.box_y + self.box_height) >= (self.frame_height - margin)
               or (self.box_y <= margin)):
                zoom_command = -1.0

        return zoom_command


class TwitchyMotorController(MotorController):

    def __init__(self,
                 pid_gains,
                 orientation,
                 example_frame):

        super().__init__(pid_gains,
                         orientation,
                         example_frame)

        self.zoom_stop_ratio = .7

    def _calc_zoom_command(self, target_lbox, x_err, y_err, zoom_command):
        """Calculate the zoom command give pan/tilt errors

        """
        if x_err != 0.0 and y_err != 0.0:
            target_bb_pixels = self.box_width * self.box_height

            # ratio of bb pixels over whole frame pixels is below a
            # threshold then zoom
            ratio = target_bb_pixels / self.total_frame_pixels
            error = 1 - ratio
            zoom_command = 0.05 * error

            # stop zoom if either dimension of bounding box is
            l_ratio = self.zoom_stop_ratio  # length ratio
            spans_much_width = self.box_width >= l_ratio * self.frame_width
            spans_much_height = self.box_height >= l_ratio * self.frame_height
            if spans_much_width or spans_much_height:
                zoom_command = -0.1

            # margin = 100
            # if((self.box_y + self.box_height) >= (self.frame_height - margin)
            #    or (self.box_y <= margin)):
            #     zoom_command = 0.0
        else:
            zoom_command = 0.0

        return zoom_command


class BouncyZoomMotorController(MotorController):

    def __init__(self,
                 pid_gains,
                 orientation,
                 example_frame):

        super().__init__(pid_gains,
                         orientation,
                         example_frame)

        self.zoom_stop_ratio = .7

    def _calc_zoom_command(self, target_lbox, x_err, y_err, zoom_command):
        """Calculate the zoom command give pan/tilt errors

        """
        if x_err != 0.0 and y_err != 0.0:
            target_bb_pixels = self.box_width * self.box_height

            # ratio of bb pixels over whole frame pixels is below a
            # threshold then zoom
            ratio = target_bb_pixels / self.total_frame_pixels
            if ratio < .1:
                zoom_command = 0.01
            elif ratio > .3:
                zoom_command = -0.01
            else:
                zoom_command = 0.0
        else:
            zoom_command = 0.0

        return zoom_command


class PtzCam():
    """Class to control PTZ on ONVIF-compliant PTZ IP Camera

    Allows control of the pan, tilt, and zoom of an ONVIF-compliant IP
    camera that has PTZ capability.

    """
    def __init__(self,
                 ip_address=None,
                 port='80',
                 user=None,
                 pword=None):
        """ PtzCam constructor

        Parameters
        ----------
        ip_address : str
           The IP address of the camera to connect to.
        port : str
           ONVIF port on the camera. This is usually 80
        user : str
           Valid username of account on the IP camera being connected to.
        pword : str
           Password for the account on the IP camera.

        """

        mycam = ONVIFCamera(ip_address, port, user, pword)
        media_service = mycam.create_media_service()
        self.ptz_service = mycam.create_ptz_service()
        self.imaging_service = mycam.create_imaging_service()

        # on hikvision cameras there have been 3 profiles: one for each stream
        self.media_profile = media_service.GetProfiles()[0]

        self.video_source = media_service.GetVideoSources()[0]

        self.pan_bounds = [-1.0, 1.0]
        self.tilt_bounds = [-1.0, 1.0]
        self.zoom_bounds = [0.0, 1.0]

        vst = {'VideoSourceToken': self.video_source.token}
        self.imaging_settings = self.imaging_service.GetImagingSettings(vst)

        min_iris = self.imaging_settings.Exposure.MinIris,
        max_iris = self.imaging_settings.Exposure.MaxIris,
        self.iris_bounds = [min_iris,
                            max_iris]

        min_exposure_time = self.imaging_settings.Exposure.MinExposureTime
        max_exposure_time = self.imaging_settings.Exposure.MaxExposureTime
        self.exposure_time_bounds = [min_exposure_time,
                                     max_exposure_time]

        # self.moverequest = self.ptz.create_type('ContinuousMove')
        # self.moverequest.ProfileToken = media_profile.token
        # # if self.moverequest.Velocity is None:
        #     # self.moverequest.Velocity = {'PanTilt': {'x': -1, 'y': -1},
        #     #                              'Zoom': {'x': 0.0}}

        # self.moverequest.Velocity = {'PanTilt': {'x': -1, 'y': 1},
        #                              'Zoom': {'x': 0.0}}

    def __del__(self):
        print('[INFO] PtzCam object deletion.')

    def get_exposure(self):
        vst = {'VideoSourceToken': self.video_source.token}
        self.imaging_settings = self.imaging_service.GetImagingSettings(vst)

        exp_time = self.imaging_settings.Exposure.ExposureTime
        gain = self.imaging_settings.Exposure.Gain
        iris = self.imaging_settings.Exposure.Iris

        return exp_time, gain, iris

    def _send_imaging_settings(self):
        command_dict = {'VideoSourceToken': self.video_source.token,
                        'ImagingSettings': self.imaging_settings}
        self.imaging_service.SetImagingSettings(command_dict)

    def set_exposure_to_auto(self):
        self.imaging_settings.Exposure['Mode'] = 'AUTO'
        self._send_imaging_settings()

    def set_focus_to_auto(self):
        self.imaging_settings.Focus['AutoFocusMode'] = 'AUTO'
        self._send_imaging_settings()

    def set_focus_to_manual(self):
        self.imaging_settings.Focus['AutoFocusMode'] = 'MANUAL'
        self._send_imaging_settings()

    def set_exposure_time(self, exposure_time):
        # need to implement bound checking using bounds gotten in constructor
        self.imaging_settings.Exposure['Mode'] = 'MANUAL'
        self.imaging_settings.Exposure['ExposureTime'] = exposure_time
        self._send_imaging_settings()

    def set_iris(self, iris):
        # need to implement bound checking using bounds gotten in constructor
        self.imaging_settings.Exposure['Mode'] = 'MANUAL'
        self.imaging_settings.Exposure['Iris'] = iris
        self._send_imaging_settings()

    def set_gain(self, gain):
        # need to implement bound checking using bounds
        self.imaging_settings.Exposure['Mode'] = 'MANUAL'
        self.imaging_settings.Exposure['Gain'] = gain
        self._send_imaging_settings()

    def focus_out(self):
        focus_request = self.imaging_service.create_type('Move')
        focus_request.VideoSourceToken = self.video_source.token
        focus_request.Focus = {'Continuous': {'Speed': 0.1}}
        self.imaging_service.Move(focus_request)

    def focus_in(self):
        focus_request = self.imaging_service.create_type('Move')
        focus_request.VideoSourceToken = self.video_source.token
        focus_request.Focus = {'Continuous': {'Speed': -0.1}}
        self.imaging_service.Move(focus_request)

    def focus_stop(self):
        self.imaging_service.Stop(self.video_source.token)

    def move(self, x_velocity, y_velocity):
        move_request = self.ptz_service.create_type('ContinuousMove')
        move_request.ProfileToken = self.media_profile.token
        move_request.Velocity = {'PanTilt': {'x': x_velocity, 'y': y_velocity},
                                 'Zoom': {'x': 0.0}}
        self.ptz_service.ContinuousMove(move_request)

    def move_w_zoom(self, x_velocity, y_velocity, zoom_command):
        x_velocity = float(_check_zeroness(x_velocity))
        y_velocity = float(_check_zeroness(y_velocity))
        zoom_command = _check_zeroness(zoom_command)

        move_request = self.ptz_service.create_type('ContinuousMove')
        move_request.ProfileToken = self.media_profile.token
        move_request.Velocity = {'PanTilt': {'x': x_velocity, 'y': y_velocity},
                                 'Zoom': {'x': zoom_command}}
        self.ptz_service.ContinuousMove(move_request)

    def _prep_abs_move(self):
        move_request = self.ptz_service.create_type('AbsoluteMove')
        move_request.ProfileToken = self.media_profile.token
        if move_request.Position is None:
            move_request.Position = self.ptz_service.GetStatus({'ProfileToken': self.media_profile.token}).Position
            move_request.Speed = self.media_profile.PTZConfiguration.DefaultPTZSpeed

        return move_request

    def get_position(self):
        move_request = self.ptz_service.create_type('AbsoluteMove')
        move_request.ProfileToken = self.media_profile.token
        position = self.ptz_service.GetStatus({'ProfileToken': self.media_profile.token}).Position

        # x = pan, y = tilt
        return (position['PanTilt']['x'],
                position['PanTilt']['y'],
                position['Zoom']['x'])

    def absmove(self, x_pos, y_pos):
        move_request = self._prep_abs_move()
        move_request.Position.PanTilt.x = x_pos
        move_request.Position.PanTilt.y = y_pos
        self.ptz_service.AbsoluteMove(move_request)

    def twitch(self):
        pan, tilt, zoom = self.get_position()

        pan_2 = np.clip(pan + 0.1, *self.pan_bounds)
        tilt_2 = np.clip(tilt + 0.2, *self.tilt_bounds)

        self.absmove_w_zoom(pan, tilt_2, zoom)
        time.sleep(1.0)
        self.absmove_w_zoom(pan_2, tilt_2, zoom)
        time.sleep(1.0)
        self.absmove_w_zoom(pan_2, tilt, zoom)
        time.sleep(1.0)
        self.absmove_w_zoom(pan, tilt, zoom)
        time.sleep(1.0)

    def absmove_w_zoom(self, pan_pos, tilt_pos, zoom_pos):
        """Move PTZ camera to an absolute pan, tilt, zoom state.

        Parameters
        ----------
        pan_pos : float
            The desired pan position expressed as a value in accepted
            range (often -1.0 to 1.0) that maps to the actual range of
            the camera (e.g. 0-350 degrees or 0-360 degrees).

        tilt_pos : float
            The desired tilt position expressed as a value in accepted
            range (often -1.0 to 1.0) that maps to the actual range of
            the camera (e.g. 0-90 degrees or -5 to 90 degrees)

        zoom_pos : float
            The desired zoom "position" expressed as a value in
            accepted range (often 0.0 to 1.0) that maps to the actual
            zoom range of the camera.

        Returns
        -------
        None

        """
        zoom_pos = _check_zeroness(zoom_pos)
        pan_pos = _check_zeroness(pan_pos)
        tilt_pos = _check_zeroness(tilt_pos)

        move_request = self._prep_abs_move()
        move_request.Position.PanTilt.x = pan_pos
        move_request.Position.PanTilt.y = tilt_pos
        move_request.Position.Zoom.x = zoom_pos
        self.ptz_service.AbsoluteMove(move_request)

    def _wait_for_done(self, pan_goal, tilt_goal, zoom_goal, close_enough=.01):
        """Note: zoom_goal finishing not implemented

        """
        pan, tilt, zoom = self.get_position()
        while (pan >= pan_goal + close_enough
               or pan <= pan_goal - close_enough
               or tilt >= tilt_goal + close_enough
               or tilt <= tilt_goal - close_enough
               or zoom >= zoom_goal + close_enough
               or zoom <= zoom_goal - close_enough):
            time.sleep(.1)
            pan, tilt, zoom = self.get_position()

    def absmove_w_zoom_waitfordone(self,
                                   pan_pos,
                                   tilt_pos,
                                   zoom_pos,
                                   close_enough=.01):
        self.absmove_w_zoom(pan_pos, tilt_pos, zoom_pos)
        self._wait_for_done(pan_pos, tilt_pos, zoom_pos, close_enough)

    def zoom_out_full(self):
        move_request = self._prep_abs_move()
        move_request.Position.Zoom.x = 0.0
        self.ptz_service.AbsoluteMove(move_request)

    def zoom_in_full(self):
        move_request = self._prep_abs_move()
        move_request.Position.Zoom.x = 1.0
        self.ptz_service.AbsoluteMove(move_request)

    def zoom(self, zoom_command):
        zoom_command = _check_zeroness(zoom_command)

        move_request = self._prep_abs_move()
        move_request.Position.Zoom.x = zoom_command
        self.ptz_service.AbsoluteMove(move_request)

    def stop(self):
        move_request = self._prep_abs_move()
        self.ptz_service.Stop({'ProfileToken': move_request.ProfileToken})


class CalmNoZoomMotorController(MotorController):

    def __init__(self,
                 pid_gains,
                 orientation,
                 example_frame,
                 zoom_pickup=.01):

        super().__init__(pid_gains,
                         orientation,
                         example_frame)

        # this class has not been updated to way things work now
        raise NotImplementedError

        self.zoom_pickup = zoom_pickup
        self.ZOOM_STOP_RATIO = .7
        self.STOP_RANGE = .1

    def _calc_command(self, err, k):
        """Override controller command method

        The meat here is that the controller stops moving axis under
        control if error is inside of some range.  This range is
        currently hardcoded but maybe should be given as an argument
        to the constructor.

        """

        if np.abs(err) < self.STOP_RANGE:
            command = 0
        else:
            command = k * err

        command = self._ensure_command_in_bounds(command)

        return command

    def _calc_zoom_command(self, x_err, y_err, zoom_command):
        """Calculate the zoom command give pan/tilt errors

        """
        zoom_command = 0.0
        
        return zoom_command



import os

from ptzipcam.camera import Camera
from ptzipcam import ui

from dnntools import neuralnetwork as nn
# from dnntools import neuralnetwork_coral as nn

from dnntools import draw


class Tracker():
    
    def __init__(self, configs):
        
        IP = configs['IP']
        USER = configs['USER']
        PWORD = configs['PASS']
        STREAM = configs['STREAM']    
        self.cam = Camera(ip=IP, user=USER, passwd=PWORD, stream=STREAM)

        frame = self.cam.get_frame()
        self.frame_width = frame.shape[1]
        self.frame_height = frame.shape[0]

        # self.total_pixels = self.frame_width * self.frame_height
        
        self.ORIENTATION = configs['ORIENTATION']

        # CV constants
        self.TRACKED_CLASS = configs['TRACKED_CLASS']
        self.CONF_THRESHOLD = configs['CONF_THRESHOLD']
        self.NMS_THRESHOLD = configs['NMS_THRESHOLD']
        INPUT_WIDTH = configs['INPUT_WIDTH']
        INPUT_HEIGHT = configs['INPUT_HEIGHT']

        
        MODEL_PATH = configs['MODEL_PATH']
        MODEL_CONFIG = os.path.join(MODEL_PATH, configs['MODEL_CONFIG_FILE'])
        MODEL_WEIGHTS = os.path.join(MODEL_PATH, configs['MODEL_WEIGHTS_FILE'])
        CLASSES_FILE = os.path.join(MODEL_PATH, configs['CLASS_NAMES_FILE'])
        self.CLASSES = nn.read_classes_from_file(CLASSES_FILE)

        self.network = nn.ObjectDetectorHandler(MODEL_CONFIG,
                                                MODEL_WEIGHTS,
                                                INPUT_WIDTH,
                                                INPUT_HEIGHT)

        self.detected_class = 'Nothing'
        
        self.x_err = 0
        self.y_err = 0
        self.target_present = False
        self.status = 'Just got going.'

        self.frames_since_last_acq = 0
        
    def get_errors(self):
        return self.x_err, self.y_err
        
    def update(self):
        raw_frame = self.cam.get_frame()
        if raw_frame is None:
            # probably should do something with logging here
            self.status = 'Frame is None.'
            return

        # maybe the orient_frame method should not be part of ui
        # module as that is the only reason I am importing the ui
        # module into this non-UI-ish module.
        raw_frame = ui.orient_frame(raw_frame, self.ORIENTATION)
        frame = raw_frame.copy()

        outs, self.inference_time = self.network.infer(frame)

        lboxes = self.network.filter_boxes(outs,
                                           frame,
                                           self.CONF_THRESHOLD,
                                           self.NMS_THRESHOLD)


        # extract the lbox with the highest confidence (that is a target type)
        highest_confidence_tracked_class = 0
        target_lbox = None
        for lbox in lboxes:
            if self.CLASSES[lbox['class_id']] in self.TRACKED_CLASS:
                if lbox['confidence'] > highest_confidence_tracked_class:
                    highest_confidence_tracked_class = lbox['confidence']
                    target_lbox = lbox

        # if there is an appropriate lbox attempt to adjust ptz cam
        self.detected_class = 'nothing detected'
        # score = 0.0
        if target_lbox:
            self.target_present = True
            self.detected_class = self.CLASSES[target_lbox['class_id']]

            self.frames_since_last_acq = 0
            xc, yc = draw.box_to_coords(target_lbox['box'],
                                        return_kind='center')
            coords = draw.box_to_coords(target_lbox['box'])
            x, y, box_width, box_height = coords
            self.x_err = self.frame_width/2 - xc
            self.y_err = self.frame_height/2 - yc

            # target_bb_pixels = box_width * box_height
            # if (target_bb_pixels / total_pixels) < .3:
            #     zoom_command += .1
            #     if zoom_command >= 1.0:
            #         zoom_command = 1.0
            #     # zoom_command = 1.0
            # else:
            #     zoom_command = 0.0

            # filling_much_of_width = box_width >= .7 * frame_width
            # filling_much_of_height = box_height >= .7 * frame_height
            # if filling_much_of_width or filling_much_of_height:
            #     zoom_command = 0.0
        else:
            self.target_present = False
            #zoom_command = 0
            self.frames_since_last_acq += 1
            if self.frames_since_last_acq > 10:
                x_err = 0
                y_err = 0

            # if frames_since_last_acq > 30:
            #     zoom_command -= .05
            #     if zoom_command <= -1.0:
            #         zoom_command = -1.0
            #     # zoom_command = -1.0




import logging
import threading
import cv2

# latest_frame = None
# latest_frame_return = None
# lo = threading.Lock()

log = logging.getLogger(__name__)

def camera_thread_function(cap, frame):
    # global latest_frame, lo, latest_frame_return
    while True:
        # with lo:
        # latest_frame_return, latest_frame = cap.read()
        ok, frame[0] = cap.read()


class Camera():

    # def __init__(self, address='udp://127.0.0.1:5000'):
    def __init__(self,
                 ip,
                 user,
                 passwd,
                 stream=3,
                 rtsp_port=554,
                 cam_brand='hikvision'):

        if cam_brand == 'axis':
            stream_string = (':'
                             + str(rtsp_port)
                             + '/axis-media/media.amp')
        elif cam_brand == 'hikvision':
            stream_string = (':'
                             + str(rtsp_port)
                             + '/Streaming/Channels/10'
                             + str(stream))
        else:
            print('[ERROR] Camera type not recognized.')
            
        address = ('rtsp://'
                   + user
                   + ':'
                   + passwd
                   + '@'
                   + ip
                   + stream_string)
        self.frame = [None]
        # self.cap = cv2.VideoCapture(address, cv2.CAP_FFMPEG)
        self.cap = cv2.VideoCapture(address)
        ok, self.frame[0] = self.cap.read()

        self.cam_thread = threading.Thread(target=camera_thread_function,
                                           args=(self.cap, self.frame))
        self.cam_thread.daemon = True
        self.cam_thread.start()

    def get_frame(self):
        # if (latest_frame_return is not None) and (latest_frame is not None):
        #     self.frame = latest_frame.copy()

        return self.frame[0]

    def get_resolution(self):
        return self.frame[0].shape[1], self.frame[0].shape[0]

    # def release(self):
    #     self.cap.release()

    def __del__(self):
        log.info('Camera object deletion')
        self.cap.release()



"""Functions for converting between different quantities

Most of these are for converting with PTZ quantities (angles in
degrees, zoom in "power") and the associated range of the cameras.

"""

import logging

log = logging.getLogger(__name__)


def degrees_to_command(degrees, full_range):
    if degrees > full_range:
        log.warning('Angle higher than full range')
        degrees = full_range
    elif degrees < 0.0:
        log.warning('Angle lower than zero')
        degrees = 0.0

    half_range = full_range/2.0
    return (degrees - half_range)/half_range


def command_to_degrees(command, full_range):
    """Convert ptz camera command to degrees

    Parameters
    ----------

    command : float
        PTZ camera axis command, usually in the range of -1.0 to 1.0
        (if not this function probably breaks)

    full_range : float
        Full range of that axis in degrees

    Returns
    -------

    The converted value (now in degrees)

    """
    half_range = full_range/2.0
    return command * half_range + half_range


def zoom_to_power(zoom, camera_zoom_power):
    return (zoom * (camera_zoom_power - 1)) + 1


def power_to_zoom(power, camera_zoom_power):
    if power < 1.0:
        log.warning('Zoom power command is less than 1.0.'
                    + ' Forcing to 1.0')
        power = 1.0

    if power > camera_zoom_power:
        log.warning('Zoom power command is greater than full zoom'
                    + ' Forcing to full zoom')
        power = camera_zoom_power

    return (power - 1)/(camera_zoom_power - 1)



"""Classes for writing outputs to video files

"""

import cv2


class DilationVideoWriter():
    """Manages writing of video with built in time dilation

    """


    def __init__(self,
                 output_file,
                 codec,
                 frame_rate,
                 frame_shape,
                 frame_window):

        self.frame_window = frame_window

        self.interval = 10
        self.latch_count = 0

        self.out = cv2.VideoWriter(output_file,
                                   codec,
                                   frame_rate,
                                   frame_shape)
        self.count = 0

    def update(self, frame, target_detected):
        """Updates the video writer

        Method to run each time through control loop that writes a
        frame to video a file based on the desired time dilation
        mechanics.

        """
        if target_detected:
            self.interval = 1
            # any time detected restart latch_count
            self.latch_count = 0

        self.latch_count += 1
        if self.latch_count > self.frame_window:
            self.interval = 10
            self.latch_count = 0

        self.count += 1
        if self.count >= self.interval:
            self.out.write(frame)
            self.count = 0

    def release(self):
        """Release VideoWriter object

        """
        self.out.release()



"""Setup.py for this project

"""

import os
import re

import setuptools


def read(filename):
    """Read file using path of this file to find it

    """
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)
    with open(path, 'r', encoding='utf-8') as readfile:
        return readfile.read()


def find_version(text):
    """Extract the version from a file

    Using regular expressions to find the string that looks like the
    version number

    """
    match = re.search(r"^__version__\s*=\s*['\"](.*)['\"]\s*$", text,
                      re.MULTILINE)
    return match.group(1)


AUTHOR = "Ian Ingram"
DESC = "Package for controlling PTZ IP cameras that support ONVIF"

setuptools.setup(
    name="ptzipcam",
    description=DESC,
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    license="MIT",
    version=find_version(read('ptzipcam/__init__.py')),
    author=AUTHOR,
    packages=['ptzipcam'],
    include_package_data=True,
    install_requires=[
        'numpy',
        'opencv-python',
        'onvif-zeep',
        'camml',
    ],
    extra_require={
        'examples': [
            'pyyaml',
            'imutils',
            'screeninfo',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Development Status :: 2 - Pre-Alpha',
        'Operating System :: POSIX :: Linux',
    ],
)



"""Tools for recording output from camera applications

At this stage, this amounts to a single class for recording stills.


"""
import logging
import os

from datetime import datetime
import cv2

log = logging.getLogger(__name__)


class ImageStreamRecorder():  # pylint: disable=too-few-public-methods
    """Handles recording images with their associate metadata

    The metadata are the PTZ state when the image was captured and
    objection detection (bounding box data, detected class, score)
    tied to that images capture.

    """

    def __init__(self, path):
        log.debug('Initialize recorder.')
        self.path = path
        self.timestamp_format = "%Y-%m-%dT%H-%M-%S_%f"
        timestamp_string = datetime.now().strftime(self.timestamp_format[:-3])
        image_folder_name = timestamp_string + '_images'
        self.image_path = os.path.join(self.path, image_folder_name)
        if not os.path.exists(self.image_path):
            os.mkdir(self.image_path)
        else:
            warning = ('[WARNING] Images path already exists so there could '
                       'end up being images from two runs in one folder or '
                       'worse images could be over-written (not too likely as '
                       'they have millisecond timestamps but not impossible.).'
                       ' This also means that it is likely a system clock is '
                       'off for the folder name (also timestamped) to be '
                       'duplicate')
            log.warning(warning)
        self.record_filename = timestamp_string + '.csv'
        self.record_filename = os.path.join(self.path, self.record_filename)
        with open(self.record_filename, 'w', encoding='utf-8') as record_file:
            header_line = ('IMAGE_FILE,PAN_ANGLE,TILT_ANGLE,'
                           'ZOOM_POWER,CLASS,SCORE,X,Y,W,H\n')
            record_file.write(header_line)

    def record_image(self,  # pylint: disable=too-many-locals
                     image,
                     ptz_state,
                     detected_class,
                     target_lbox):
        """Record a single image to a file and append the associated PTZ state
        and object detection data to the recording CSV file.

        Parameters
        ----------

        image :
            The image to record

        ptz_state : 3-tuple
            Contains the pan, tilt, zoom of the camera at the time the
            image was captured.

        detected_class :
            The name of the detected class

        target_lbox : dict
            The parameters of the labeled bounding box of the detected target.

        """

        pan_angle, tilt_angle, zoom = ptz_state

        front_bit = datetime.now().strftime(self.timestamp_format)[:-3]
        image_filename = front_bit + '.jpg'

        # full_path = os.path.join(self.path, 'images')
        # if not os.path.exists(full_path):
        #     os.mkdir(full_path)

        image_filename_w_path = os.path.join(self.image_path,
                                             image_filename)
        cv2.imwrite(image_filename_w_path, image)

        if target_lbox:
            score = 100 * target_lbox['confidence']
            box = target_lbox['box']
            box_x_coord, box_y_coord, box_width, box_height = box
        else:
            score = 0
            box_x_coord = 0
            box_y_coord = 0
            box_width = 0
            box_height = 0

        strg = '{},{:.2f},{:.2f},{:.2f},{},{:.1f},{},{},{},{}\n'
        record_line = strg.format(image_filename,
                                  pan_angle,
                                  tilt_angle,
                                  zoom,
                                  detected_class,
                                  score,
                                  box_x_coord,
                                  box_y_coord,
                                  box_width,
                                  box_height)

        with open(self.record_filename, 'a', encoding='utf-8') as record_file:
            record_file.write(record_line)
